/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT (c) 2008 - 2013 SRON (R.M.van.Hees@sron.nl)

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License, version 2, as
   published by the Free Software Foundation.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, 
   Boston, MA  02111-1307, USA.

.IDENTifer   NADC_RD_MCFS
.AUTHOR      R.M. van Hees
.KEYWORDS    SCIA Ife/MCFS
.LANGUAGE    ANSI C
.PURPOSE     read Ife MCFS products
.INPUT/OUTPUT
  call as   nrec = NADC_RD_MCFS( flname, &hdr, &rec );
     input:  
             char   mcfs_name[]    : filename of MCFS product
    output:  
             struct mcfs_hdr *hdr  : header of MCFS product
             struct mcfs_rec **rec : structure holding MCFS tile information

.RETURNS     number of records read (unsigned int)
.COMMENTS    None
.ENVIRONment None
.VERSION     1.0     02-Jul-2008   initial release by R. M. van Hees
------------------------------------------------------------*/
/*
 * This code needs the GNU version of basename (not POSIX)
 */
#define  _GNU_SOURCE

/*+++++ System headers +++++*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <zlib.h>

/*+++++ Local Headers +++++*/
#include <defs_nadc.h>
#include <defs_mcfs.h>

/*+++++ Macros +++++*/
#define MAX_LINE_LENGTH  256

#define NINT(a) ((a) >= 0.f ? (int)((a)+0.5) : (int)((a)-0.5))

/*+++++ Global Variables +++++*/
	/* NONE */

/*+++++ Static Variables +++++*/
	/* NONE */

/*+++++++++++++++++++++++++ Static Functions +++++++++++++++++++++++*/
/*+++++++++++++++++++++++++
.IDENTifer   Read_MCFS_Header
.PURPOSE     read header of (ascii) Ife MCFS product
.INPUT/OUTPUT
  call as   Read_MCFS_Header( flname, hdr );
     input:
            char flname[]        :  name of the MCFS product
    output:
            struct mcfs_hdr *hdr :  header info from MCFS product

.RETURNS     nothing, error status is passed by global nadc_stat
.COMMENTS    static function
-------------------------*/
static
void Read_MCFS_Header( const char flname[], /*@out@*/ struct mcfs_hdr *hdr )
   /*@globals  errno, nadc_stat, nadc_err_stack;@*/
   /*@modifies errno, nadc_stat, nadc_err_stack, hdr@*/
{
     const char prognm[] = "Read_MCFS_Header";

     char   *pntr, ctemp[SHORT_STRING_LENGTH];

     char   line[MAX_LINE_LENGTH];

     unsigned short nline = 0;

     gzFile fp;
/*
 * open the MCFS product
 */
     if ( (fp = gzopen( flname, "r" )) == NULL )
	  NADC_RETURN_ERROR( prognm, NADC_ERR_FILE, flname );

     (void) strlcpy( ctemp, basename( flname ), SHORT_STRING_LENGTH );
     if ( (pntr = strstr( ctemp, ".gz" )) != NULL ) *pntr = '\0';
     (void) strlcpy( hdr->product, ctemp, ENVI_FILENAME_SIZE+6 );
     if ( (pntr = strstr( ctemp, ".txt" )) != NULL ) *pntr = '\0';
     (void) strcat( ctemp, ".N1" );
     (void) strlcpy( hdr->l1b_product, ctemp+5, ENVI_FILENAME_SIZE );
     (void) strlcpy( hdr->institution, "IUP-IFE", 8 );
     (void) strlcpy( hdr->contact, "alexk@iup.physik.uni-bremen.de", 64 );
     (void) strlcpy( hdr->creation_date, "01-JUL-2008 00:00:00", 
		     MAX_UTC_STRING );
     NADC_RECEIVEDATE( flname, hdr->receive_date );
     hdr->file_size = NADC_FILESIZE( flname );

     do {
          if ( gzeof ( fp ) == 1 ) break;
          if ( gzgets( fp, line, MAX_LINE_LENGTH ) == NULL ) break;
	  if ( strstr( line, "*****" ) != NULL ) continue;

	  ++nline;
     } while( 1 == 1 );

     hdr->numRec = nline;
     (void) gzclose( fp );
}

/*+++++++++++++++++++++++++
.IDENTifer   Read_MCFS_Record
.PURPOSE     read data records from (ascii) Ife MCFS products
.INPUT/OUTPUT
  call as   nline = Read_MCFS_Record( fp, rec );
     input:
            char flname[]        :  name of the MCFS product
    output:
            struct mcfs_rec *rec :  data records from the MCFS product

.RETURNS     number of lines erad from product (unsigned short)
.COMMENTS    static function
-------------------------*/
static 
unsigned short Read_MCFS_Record( const char flname[], 
			       /*@out@*/ struct mcfs_rec *rec )
   /*@globals  errno, nadc_stat, nadc_err_stack;@*/
   /*@modifies errno, nadc_stat, nadc_err_stack, rec@*/
{
     const char prognm[] = "Read_MCFS_Record";

     gzFile fp = NULL;

     char           line[MAX_LINE_LENGTH], cbuff[SCIA_DATE_LENGTH];
     unsigned short nline = 0;
     int            mjd2000;
     unsigned int   second, mu_sec;
     float          intg_time, lon_cntr, lat_cntr;

     const char fmt_rec[] = "%f %f %f %f %hu %hu %hhu %f %f %f %f %f %s %s %u";
     const double SecPerDay = 60. * 60 * 24;
/*
 * open the MCFS product
 */
     if ( (fp = gzopen( flname, "r" )) == NULL )
	  NADC_GOTO_ERROR( prognm, NADC_ERR_FILE, flname );

     do {
          if ( gzeof ( fp ) == 1 ) goto done;
          if ( gzgets( fp, line, MAX_LINE_LENGTH ) == NULL ) goto done;
	  if ( strstr( line, "*****" ) != NULL ) continue;

	  (void) sscanf( line, fmt_rec, &lat_cntr, &lon_cntr, 
			 &rec->cloudFractionThick, &rec->cloudFraction,
			 &rec->meta.numMeris, &rec->meta.numPMD, 
			 &rec->meta.quality, &rec->meta.sun_zen_ang, 
			 &rec->meta.sun_azi_ang, &rec->meta.sat_zen_ang, 
			 &rec->meta.sat_azi_ang, &intg_time, rec->time, 
			 rec->date, &rec->orbit );
	  (void) snprintf( cbuff, SCIA_DATE_LENGTH, "%s %s",
			   rec->date, rec->time );
	  if ( rec->meta.numMeris > (unsigned short) (14400 * intg_time) )
	       rec->meta.numMeris = SHRT_MAX;
	  ASCII_2_MJD( cbuff, &mjd2000, &second, &mu_sec );
	  rec->jday = mjd2000 + (second + mu_sec / 1e6) / SecPerDay;
	  rec->meta.intg_time = (unsigned char) NINT( 16.f * intg_time );
	  rec->center.lon = NINT( 1e6 * lon_cntr );
	  rec->center.lat = NINT( 1e6 * lat_cntr );
	  ++rec;
	  ++nline;
     } while( 1 == 1 );
 done:
     if ( fp != NULL ) (void) gzclose( fp );
     return nline;
}

/*+++++++++++++++++++++++++ Main Program or Function +++++++++++++++*/
void NADC_RD_MCFS( const char *flname, struct mcfs_hdr *hdr,
		   struct mcfs_rec **mcfs_out )
{
     const char prognm[] = "NADC_RD_MCFS";

     unsigned short numRec = 0u;

     struct mcfs_rec *rec = NULL;
/*
 * read header of the MCFS product
 */
     mcfs_out[0] = NULL;
     Read_MCFS_Header( flname, hdr );
     if ( IS_ERR_STAT_FATAL )
          NADC_RETURN_ERROR( prognm, NADC_ERR_FATAL, "Read_MCFS_Header" );
/*
 * read records from the MCFS product
 */
     rec = (struct mcfs_rec *) 
	  malloc( hdr->numRec * sizeof( struct mcfs_rec ));
     if ( rec == NULL ) NADC_RETURN_ERROR( prognm, NADC_ERR_ALLOC, "rec" );
     if ( (numRec = Read_MCFS_Record( flname, rec )) != hdr->numRec ) {
          char msg[SHORT_STRING_LENGTH];

          free( rec );
          (void) snprintf( msg, SHORT_STRING_LENGTH,
                           "failed to read all records %hu out-of %hu\n",
                           numRec, hdr->numRec );
          NADC_RETURN_ERROR( prognm, NADC_ERR_FILE_RD, msg );
     }
/*
 * fill empty variables in header structure
 */
     hdr->orbit = rec[numRec/2].orbit;
     (void) snprintf( hdr->product_datetime_start, MAX_UTC_STRING,
		      "%s %s", rec[0].date, rec[0].time );
     (void) snprintf( hdr->product_datetime_end, MAX_UTC_STRING,
		      "%s %s", rec[numRec-1].date, rec[numRec-1].time );
     mcfs_out[0] = rec;
}
