/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT (c) 2008 - 2013 SRON (R.M.van.Hees@sron.nl)

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License, version 2, as
   published by the Free Software Foundation.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, 
   Boston, MA  02111-1307, USA.

.IDENTifer   NADC_WR_NC_MCFS
.AUTHOR      R.M. van Hees
.KEYWORDS    MCFS Sciamachy Meris
.LANGUAGE    ANSI C
.PURPOSE     write MCFS product in ADAGUC format
.COMMENTS    contains NADC_MCFS_WR_NC_META and NADC_MCFS_WR_NC_REC
.ENVIRONment None
.VERSION     1.0     20-Oct-2008   initial release by R. M. van Hees
------------------------------------------------------------*/
/*
 * Define _POSIX_SOURCE to indicate
 * that this is a POSIX program
 */
#define  _POSIX_SOURCE 2

/*+++++ System headers +++++*/
#include <stdlib.h>
#include <string.h>

#ifdef _WITH_NC4
#include <hdf5.h>
#include <netcdf.h>

/*+++++ Local Headers +++++*/
#include <defs_nadc.h>
#define __NEED_ISO_ENTRIES
#include <defs_mcfs.h>

/*+++++ Macros +++++*/
	/* NONE */

/*+++++ Global Variables +++++*/
	/* NONE */

/*+++++ Static Variables +++++*/
	/* NONE */

/*+++++++++++++++++++++++++ Static Functions +++++++++++++++++++++++*/
/*+++++++++++++++++++++++++
.IDENTifer   DEFINE_NC_VAR
.PURPOSE     define a netCDF variable in ADAGUC format
.INPUT/OUTPUT
  call as   retval = DEFINE_NC_VAR( int ncid, nc_type xtype, int dimids,
                   const char *name, const char *unit, const char *long_name )
     input:
            int ncid        :   netCDF file ID
            nc_type xtype   :   netCDF variable type
	    int dimids      :   dimension IDs
	    char *name      :   short name for variable
	    char *unit      :   unit of variable (NULL if not defined)
	    char *long_name :   long name for variable

.RETURNS     Nothing
.COMMENTS    None
-------------------------*/
static
int DEFINE_NC_VAR( int ncid, nc_type xtype, int dimids,
                   const char *name, const char *unit, const char *long_name )
{
     const char prognm[] = "DEFINE_NC_VAR";

     int retval;
     int var_id;

     retval = nc_def_var( ncid, name, xtype, 1, &dimids, &var_id );
     if ( retval != NC_NOERR )
          NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     retval = nc_put_att_text( ncid, var_id, "long_name",
                               strlen(long_name), long_name );
     if ( retval != NC_NOERR )
          NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     if ( unit != NULL ) {
          retval = nc_put_att_text( ncid, var_id, "units",
                                    strlen(unit), unit );
          if ( retval != NC_NOERR )
               NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     }
     retval = nc_put_att_text( ncid, var_id, "coordinates", 7, "lon lat" );
     if ( retval != NC_NOERR )
          NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     retval = nc_put_att_text( ncid, var_id, "grid_mapping",10, "projection" );
     if ( retval != NC_NOERR )
          NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );

     return var_id;
 done:
     return -1;
}

/*+++++++++++++++++++++++++ Main Program or Function +++++++++++++++*/
/*+++++++++++++++++++++++++
.IDENTifer   NADC_MCFS_WR_NC_META
.PURPOSE     add header to MCFS product in netCDF-4 format (ADAGUC standard)
.INPUT/OUTPUT
  call as   NADC_MCFS_WR_NC_META( ncid, hdr );
     input:
            int ncid             :  netCDF file ID
	    struct mcfs_hdr *hdr :  MCFS header

.RETURNS     Nothing
.COMMENTS    static function
-------------------------*/
void NADC_MCFS_WR_NC_META( int ncid, const struct mcfs_hdr *hdr )
{
     const char prognm[] = "NADC_MCFS_WR_NC_META";

     register unsigned short ni;

     int  retval;
     int  var_id;
/*
 * Product root-data
 */
     for ( ni = 0; ni < numRootKeys; ni++ ) {
          retval = nc_put_att_text( ncid, NC_GLOBAL,
                                    meta_root_list[ni].attr_name,
                                    strlen(meta_root_list[ni].attr_value)+1,
                                    meta_root_list[ni].attr_value );
          if ( retval != NC_NOERR )
               NADC_RETURN_ERROR(prognm, NADC_ERR_FATAL, nc_strerror(retval));
     }
/*
 * Product meta-data
 */
     retval = nc_def_var( ncid, "product", NC_CHAR, 0, NULL, &var_id );
     if ( retval != NC_NOERR )
          NADC_RETURN_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     for ( ni = 0; ni < numProdKeys; ni++ ) {
	  if ( strcmp(meta_prod_list[ni].attr_name, "input_products")  == 0 )
	       meta_prod_list[ni].attr_value = hdr->l1b_product;
	  else if ( strcmp(meta_prod_list[ni].attr_name, "creation_date")  == 0 )
	       meta_prod_list[ni].attr_value = hdr->creation_date;
	  else if ( strcmp(meta_prod_list[ni].attr_name, "validity_start")  == 0 )
	       meta_prod_list[ni].attr_value = hdr->product_datetime_start;
	  else if ( strcmp(meta_prod_list[ni].attr_name, "validity_stop")  == 0 )
	       meta_prod_list[ni].attr_value = hdr->product_datetime_end;

	  retval = nc_put_att_text( ncid, var_id,
				    meta_prod_list[ni].attr_name,
				    strlen(meta_prod_list[ni].attr_value),
				    meta_prod_list[ni].attr_value );
	  if ( retval != NC_NOERR )
	       NADC_RETURN_ERROR(prognm, NADC_ERR_FATAL, nc_strerror(retval));
     }
/*
 * Custom meta-data
 */
     retval = nc_def_var( ncid, "custom", NC_CHAR, 0, NULL, &var_id );
     if ( retval != NC_NOERR )
	  NADC_RETURN_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
/*
 * Projection meta-data
 */
     retval = nc_def_var( ncid, "projection", NC_CHAR, 0, NULL, &var_id );
     if ( retval != NC_NOERR )
          NADC_RETURN_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     for ( ni = 0; ni < numGeoKeys; ni++ ) {
	  retval = nc_put_att_text( ncid, var_id,
				    geo_prod_list[ni].attr_name,
				    strlen(geo_prod_list[ni].attr_value),
				    geo_prod_list[ni].attr_value );
	  if ( retval != NC_NOERR )
	       NADC_RETURN_ERROR(prognm, NADC_ERR_FATAL, nc_strerror(retval));
     }
/*
 * ISO meta-data
 */
     retval = nc_def_var( ncid, "iso_dataset", NC_CHAR, 0, NULL, &var_id );
     if ( retval != NC_NOERR )
	  NADC_RETURN_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     for ( ni = 0; ni < numIsoKeys; ni++ ) {
	  if ( strcmp(iso_prod_list[ni].attr_name, "max-x")  == 0 || 
	       strcmp(iso_prod_list[ni].attr_name, "min-x")  == 0 ||
	       strcmp(iso_prod_list[ni].attr_name, "max-y")  == 0 ||
	       strcmp(iso_prod_list[ni].attr_name, "min-y")  == 0 ) {
	       /* these 4 vars should we written out as NC floats */
	       float fValue = atof(iso_prod_list[ni].attr_value);
	       retval = nc_put_att_float( ncid, var_id,
					  iso_prod_list[ni].attr_name, 
					  NC_FLOAT, 1, &fValue );
	  } else {
	       retval = nc_put_att_text( ncid, var_id,
					 iso_prod_list[ni].attr_name,
					 strlen(iso_prod_list[ni].attr_value)+1,
					 iso_prod_list[ni].attr_value );
	  }
	  if ( retval != NC_NOERR )
	       NADC_RETURN_ERROR(prognm, NADC_ERR_FATAL, nc_strerror(retval));
     }
}

/*+++++++++++++++++++++++++
.IDENTifer   NADC_MCFS_WR_NC_REC
.PURPOSE     write SRON MCFS records to netCDF product (ADAGUC standard)
.INPUT/OUTPUT
  call as   NADC_MCFS_WR_NC_REC( ncid, numRec, rec );
     input:
            int ncid              :  netCDF file ID
	    unsigned short numRec :  number of MCFS records
	    struct mcfs_rec *rec  :  MCFS records

.RETURNS     Nothing
.COMMENTS    static function
-------------------------*/
void NADC_MCFS_WR_NC_REC( int ncid, unsigned short numRec,
			  const struct mcfs_rec *rec )
{
     const char prognm[] = "NADC_MCFS_WR_NC_REC";

     register unsigned int nr;

     int    retval;
     int    time_id, nv_id;
     int    meta_id, var_id;
     size_t indx;

     float  *rbuff = NULL;
     double *dbuff = NULL;
     
     if ( numRec == 0 ) return;
/*
 * write dimension scale "time"
 */
     dbuff = (double *) malloc( numRec * sizeof(double) );
     if ( dbuff == NULL ) NADC_RETURN_ERROR( prognm, NADC_ERR_ALLOC, "dbuff" );

     retval = nc_def_dim( ncid, "time", (size_t) numRec, &time_id );
     if ( retval != NC_NOERR )
	  NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     retval = nc_def_var( ncid, "time", NC_DOUBLE, 1, &time_id, &var_id );
     retval = nc_put_att_text( ncid, var_id, "long_name", 4, "time" );
     retval = nc_put_att_text( ncid, var_id, "units", 34, 
			       "days since 2000-01-01 00:00:00 UTC" );
     retval = nc_put_att_text( ncid, var_id, "calendar", 4, "none" );
     for ( nr = 0; nr < numRec; nr++ ) dbuff[nr] = rec[nr].jday;
     retval = nc_put_var_double( ncid, var_id, dbuff );

     retval = nc_def_dim( ncid, "nv", NUM_CORNERS, &nv_id );
     if ( retval != NC_NOERR )
	  NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
/*
 * write longitude and latitude of measurements
 */
     rbuff = (float *) malloc( numRec * sizeof(float) );
     if ( rbuff == NULL ) NADC_RETURN_ERROR( prognm, NADC_ERR_ALLOC, "rbuff" );

     retval = nc_def_var( ncid, "lon", NC_FLOAT, 1, &time_id, &var_id );
     if ( retval != NC_NOERR )
	  NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     retval = nc_put_att_text( ncid, var_id, "long_name", 9, "longitude" );
     retval = nc_put_att_text( ncid, var_id, "units", 12, "degrees_east" );
     retval = nc_put_att_text( ncid, var_id, "standard_name", 9, "longitude" );
     for ( nr = 0; nr < numRec; nr++ ) rbuff[nr] = rec[nr].center.lon / 1e6;
     retval = nc_put_var_float( ncid, var_id, rbuff );

     retval = nc_def_var( ncid, "lat", NC_FLOAT, 1, &time_id, &var_id );
     if ( retval != NC_NOERR )
	  NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
     retval = nc_put_att_text( ncid, var_id, "long_name", 8, "latitude" );
     retval = nc_put_att_text( ncid, var_id, "units", 13, "degrees_north" );
     retval = nc_put_att_text( ncid, var_id, "standard_name", 8, "latitude" );
     for ( nr = 0; nr < numRec; nr++ ) rbuff[nr] = rec[nr].center.lat / 1e6;
     retval = nc_put_var_float( ncid, var_id, rbuff );
/*
 * write pixel meta-data as compound dataset
 */
     retval = nc_def_compound( ncid, sizeof(struct mcfs_meta_rec), 
			       "meta_rec", &meta_id );
     nc_insert_compound( ncid, meta_id, "integration_time",
			 HOFFSET(struct mcfs_meta_rec, intg_time ), NC_UBYTE );
     nc_insert_compound( ncid, meta_id, "quality_of_retrieval",
			 HOFFSET(struct mcfs_meta_rec, quality ), NC_UBYTE );
     nc_insert_compound( ncid, meta_id, "meris_in_scia_pixel",
			 HOFFSET(struct mcfs_meta_rec, numMeris ), NC_USHORT );
     nc_insert_compound( ncid, meta_id, "number_of_PMDs",
			 HOFFSET(struct mcfs_meta_rec, numPMD ), NC_USHORT );

     nc_insert_compound( ncid, meta_id, "solar_zenith_angle",
			 HOFFSET( struct mcfs_meta_rec, sun_zen_ang ), 
			 NC_FLOAT );
     nc_insert_compound( ncid, meta_id, "solar_azimuth_angle",
			 HOFFSET( struct mcfs_meta_rec, sun_azi_ang ), 
			 NC_FLOAT );
     nc_insert_compound( ncid, meta_id, "satellite_zenith_angle",
			 HOFFSET( struct mcfs_meta_rec, sat_zen_ang ), 
			 NC_FLOAT );
     nc_insert_compound( ncid, meta_id, "satellite_azimuth_angle",
			 HOFFSET( struct mcfs_meta_rec, sat_azi_ang ), 
			 NC_FLOAT );
     nc_def_var( ncid, "tile", meta_id, 1, &time_id, &var_id );
     retval = nc_put_att_text( ncid, var_id, "long_name", 36, 
			       "pixel_properties_and_retrieval_flags" );
     for ( indx = 0; indx < (size_t) numRec; indx++ ) {
	  retval = nc_put_var1( ncid, var_id, &indx, &rec[indx].meta );
     }
/*
 * write datasets (cloudFraction,cloudFractionThick)
 */
     var_id = DEFINE_NC_VAR( ncid, NC_FLOAT, time_id, "cloudFraction", NULL,
			     "cloud_area_fraction" );
     if ( IS_ERR_STAT_FATAL ) 
	  NADC_GOTO_ERROR( prognm, NADC_ERR_HDF_WR, "cloudFraction" );
     for ( nr = 0; nr < numRec; nr++ ) rbuff[nr] = rec[nr].cloudFraction;
     retval = nc_put_var_float( ncid, var_id, rbuff );
     var_id = DEFINE_NC_VAR( ncid, NC_FLOAT, time_id, "cloudFractionThick", 
			     NULL, "thick_cloud_area_fraction" );
     if ( IS_ERR_STAT_FATAL ) 
	  NADC_GOTO_ERROR( prognm, NADC_ERR_HDF_WR, "cloudFractionThick" );
     for ( nr = 0; nr < numRec; nr++ ) rbuff[nr] = rec[nr].cloudFractionThick;
     retval = nc_put_var_float( ncid, var_id, rbuff );

 done:
     if ( rbuff != NULL ) free( rbuff );
     if ( dbuff != NULL ) free( dbuff );
}
#endif   /* _WITH_NC4 */
