/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT (c) 2001 - 2013 SRON (R.M.van.Hees@sron.nl)

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License, version 2, as
   published by the Free Software Foundation.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, 
   Boston, MA  02111-1307, USA.

.IDENTifer   NADC_CodeCalib_GOME
.AUTHOR      R.M. van Hees
.KEYWORDS    calibration option encoding/decoding
.LANGUAGE    ANSI C
.PURPOSE     GOME level 1b calibration parameter coding
.INPUT/OUTPUT
   call as:  calib_mask = NADC_GOME_CalibMask( calib_str );
     input:
            char calib_str[] :           string with encoded calibration steps

.RETURNS     calibration mask (unsigned short)
.INPUT/OUTPUT
   call as:  nr_char = NADC_GOME_CalibStr( calib_val, calib_str );
     input:
            unsigned short calib_val :   encoded calibration steps
    output:  
            char calib_str[] :           string with encoded calibration steps

.RETURNS     size of the calib string
.INPUT/OUTPUT
  call as:   NADC_GOME_CalibShow( FILE *stream );
    input:
            FILE stream :   stream to show available calibration options

.COMMENTS    None
.ENVIRONment None
.VERSION      2.0   31-Jan-2005	complete rewrite, and combined functionality
                                nadc_encodecalib.c/nadc_decodecalib.c, RvH
              1.0   04-Nov-2003	created by R. M. van Hees 
------------------------------------------------------------*/
/*
 * Define _POSIX_SOURCE to indicate
 * that this is a POSIX program
 */
#define  _POSIX_SOURCE 2

/*+++++ System headers +++++*/
#include <stdio.h>
#include <string.h>

/*+++++ Local Headers +++++*/
#include <defs_nadc.h>

/*+++++ Static Variables +++++*/
static 
const struct calib_opt {
     const char key[2];
     const unsigned short val;
     const char *desc;
} calib_opts[] = {
     { "L", GOME_CAL_LEAK, "correct for fixed-pattern noise and leakage" },
     { "A", GOME_CAL_FPA, "correct band 1a for cross-talk" },
     { "F", GOME_CAL_FIXED, "correct signals for pixel-to-pixel variations" },
     { "S", GOME_CAL_STRAY, "correct signals for stray-light" },
     { "N", GOME_CAL_NORM, "normalise signals to 1 second integration time" },
     { "P", GOME_CAL_POLAR, "Polarisation sensitivity correction" },
     { "I", GOME_CAL_INTENS, "absolute radiance calibration " },
     { "B", GOME_CAL_BSDF, "divide signals from Sun measurements by BSDF" },
     { "a", GOME_CAL_ALBEDO, "convert intensity to albedo" },
     { "j", GOME_CAL_JUMPS, "not implemented" },
     { "d", GOME_CAL_AGING, "not implemented" },
     { "U", GOME_CAL_UNIT, "unit conversion to photons/(s~cm$^{2}$sr~nm)" }
};

static
const int NumOptions = sizeof( calib_opts ) / sizeof( struct calib_opt );

/*+++++++++++++++++++++++++ Main Program or Function +++++++++++++++*/
unsigned short NADC_GOME_CalibMask( const char calib_str[] )
{
     register int nr;

     unsigned short calib_mask = CALIB_NONE;
/*
 * first check the easy declarations
 */
     if ( strlen( calib_str ) == 0 || strncmp( calib_str, "none", 4 ) == 0 )
	  return calib_mask;
/*
 * check every available calibration option
 */
     for ( nr = 0; nr < NumOptions; nr++ ) {
	  if ( strstr( calib_str, calib_opts[nr].key ) != NULL )
	       calib_mask |= calib_opts[nr].val;
     }
     return calib_mask;
}

unsigned short NADC_GOME_CalibStr( unsigned short calib_val, char calib_str[] )
{
     register int nr;

     (void) strcpy( calib_str, "" );
     for ( nr = 0; nr < NumOptions; nr++ ) {
	  if ( (calib_val & calib_opts[nr].val) != USHRT_ZERO ) {
	       if ( *calib_str != '\0' ) (void) strcat( calib_str, "," );
	       (void) strcat( calib_str, calib_opts[nr].key );
	  }
     }
     return (unsigned short) strlen( calib_str );
}

void NADC_GOME_CalibShow( FILE *stream )
{
     register int nr;

     for ( nr = 0; nr < NumOptions; nr++ ) {
	  (void) fprintf( stream, "   %s:\t%s\n", calib_opts[nr].key, 
			  calib_opts[nr].desc );
     }
}
